import argparse
import collections
import dataclasses
import re
import sys

import bs4
import requests
from prometheus_client import Gauge

from . import Exporter, Metric

from typing import cast, Any, Counter, Dict, Iterator, List

JSONType = Dict[str, Any]


@dataclasses.dataclass
class JenkinsConnection:
    url: str
    username: str
    password: str


@dataclasses.dataclass
class NodeInfo:
    api: Dict[str, Any]
    config: bs4.Tag


@dataclasses.dataclass
class NodesInfo:
    api: List[Dict[str, Any]]
    nodes: Dict[str, NodeInfo]


class JenkinsExporter(Exporter):
    def parse_args(self) -> argparse.Namespace:
        options = super(JenkinsExporter, self).parse_args()

        self.jenkins = JenkinsConnection(
            url=self.config['jenkins']['master_url'],
            username=self.config['jenkins']['username'],
            password=self.config['jenkins']['password']
        )

        return options

    def _fetch_json(self, url: str) -> JSONType:
        return cast(
            JSONType,
            requests.get(
                '{}/{}'.format(self.jenkins.url, url),
                auth=(self.jenkins.username, self.jenkins.password)
            ).json()
        )

    def _fetch_xml(self, url: str) -> bs4.BeautifulSoup:
        return bs4.BeautifulSoup(
            requests.get(
                '{}/{}'.format(self.jenkins.url, url),
                auth=(self.jenkins.username, self.jenkins.password)
            ).content,
            'xml'
        )

    def fetch_data(self) -> None:
        self.data_nodes = NodesInfo(
            api=[],
            nodes={}
        )

        # Fetch summary data we can get for all nodes via JSON API.
        self.data_nodes.api = cast(
            List[Dict[str, Any]],
            self._fetch_json('/computer/api/json?depth=1')['computer']
        )

        # For each node, fetch its configuration, and extract its JSON API data from the global structure.
        for node_info in self.data_nodes.api:
            display_name = node_info['displayName']

            if display_name == 'master':
                continue

            self.data_nodes.nodes[display_name] = NodeInfo(
                api=node_info,
                config=self._fetch_xml('/computer/{}/config.xml'.format(display_name))
            )

        self.data_queue = self._fetch_json('/queue/api/json')

    def _create_label_count_metric(self) -> Metric:
        labels: Counter[str] = collections.Counter()

        for display_name, node in self.data_nodes.nodes.items():
            node_labels_el = node.config.find('label')
            if node_labels_el is None:
                print('Node {} has no "label" configuration'.format(display_name), file=sys.stderr)
                continue

            for label in node_labels_el.text.split(' '):
                labels[label] += 1

        # Make sure to set default value for labels we haven't seen.
        for expected_label in self.config['expected-labels']:
            if expected_label in labels:
                continue

            labels[expected_label] = 0

        metric = Gauge(
            'node_labels',
            'Number of executors having a given label',
            ['label']
        )

        for node_label, count in labels.items():
            metric.labels(
                label=node_label
            ).set(count)

        return metric

    def _create_build_count_metric(self) -> Metric:
        running_jobs: Counter[str] = collections.Counter()
        queued_jobs: Counter[str] = collections.Counter()

        # Running builds first
        for display_name, node in self.data_nodes.nodes.items():
            for executor in node.api['executors']:
                if executor['currentExecutable'] is None:
                    continue

                if 'url' not in executor['currentExecutable']:
                    print('Executor {}:{} has no build URL'.format(display_name, executor['number']), file=sys.stderr)
                    continue

                match = re.match(r'{}/job/(.+?)/\d+/'.format(self.jenkins.url), executor['currentExecutable']['url'])
                if match:
                    running_jobs[match.group(1)] += 1

        # Queued builds next
        for queue_item in self.data_queue['items']:
            if queue_item['_class'] != 'hudson.model.Queue$BuildableItem':
                continue

            queued_jobs[queue_item['task']['name']] += 1

        # Make sure to set default value for jobs we haven't seen.
        for expected_job in self.config['expected-jobs']:
            if expected_job not in running_jobs:
                running_jobs[expected_job] = 0

            if expected_job not in queued_jobs:
                queued_jobs[expected_job] = 0

        metric = Gauge(
            'build_count',
            'Number of builds of a given job',
            ['job', 'state']
        )

        for job_name, count in running_jobs.items():
            metric.labels(
                job=job_name,
                state='running'
            ).set(count)

        for job_name, count in queued_jobs.items():
            metric.labels(
                job=job_name,
                state='queued'
            ).set(count)

        return metric

    def create_metric(self) -> Iterator[Metric]:
        yield self._create_label_count_metric()
        yield self._create_build_count_metric()


def main() -> None:
    """
    Fetches metrics from Jenkins and exports them to Prometheus format.
    """

    JenkinsExporter().run()


if __name__ == "__main__":
    main()



import json
import sys
import subprocess

from . import Exporter, Metric

from prometheus_client import Gauge

from typing import Any, Iterator


class OpenstackExporter(Exporter):
    def fetch_data(self) -> None:
        """
        Fetch a openstack data from server. ``self.data`` would become a structure like this:

          [{'Name': 'nameOfData', 'Value': 1024}, ...]
        """

        try:
            self.data = json.loads(
                subprocess.check_output(
                    "{0} limits show -f json --absolute --reserved".format(
                        self.get_command_openstack()
                    ),
                    shell=True
                )
            )
        except (subprocess.CalledProcessError, KeyError) as e:
            sys.stderr.write(
                "[ERROR] Phoebe Openstack: Can not fetch Openstack data.\n{0}{1}\n\n".format(type(e), e.args)
            )
            raise

    def get_command_openstack(self) -> str:
        """
        It uses credentials from settings to create an Openstack client command-line.

        :returns: Shell command to login to Openstack
        """
        return """
            openstack --os-auth-url={self.config[client][os_auth_url]} \
                        --os-identity-api-version={self.config[client][os_identity_api_version]} \
                        --os-user-domain-name={self.config[client][os_user_domain_name]} \
                        --os-project-domain-name={self.config[client][os_project_domain_name]} \
                        --os-project-name={self.config[client][os_project_name]} \
                        --os-username={self.config[client][os_username]} \
                        --os-password={self.config[client][os_password]} \
            """.format(self=self)

    def extract_value(self, name: str) -> Any:
        """
        It extract value from a Openstack statistical data.

        :param name str: A name of specific value
        :returns: A value of a given name
        """
        return list(filter(lambda d: d['Name'] == name, self.data))[0]['Value']

    def create_metric(self) -> Iterator[Metric]:
        """
        Generate Prometheus data for metrics. Definition of each metric is read from the configuration
        and yielded to the caller.
        """

        for metric_spes in self.config['metrics']:
            prom_metric = Gauge(
                metric_spes['name'], metric_spes['help'], ['type'],
            )

            for metric_label in metric_spes['labels']:
                prom_metric.labels(type=metric_label['type']).set(
                    self.extract_value(metric_label['os-property'])
                )

            yield prom_metric


def main() -> None:
    """
    Fetches metrics from Openstack and exports them to Prometheus format.
    """

    OpenstackExporter().run()


if __name__ == "__main__":
    main()



import argparse
import sys
import traceback

from prometheus_client import CollectorRegistry, generate_latest
from prometheus_client.metrics import MetricWrapperBase as Metric
import ruamel.yaml

from typing import Any, Dict, Iterator

SettingsType = Dict[str, Any]


class Exporter:
    def create_args_parser(self) -> argparse.ArgumentParser:
        """
        Create and return a parser for command-line options. The parser is already set to accept
        common, shared options ``--config`` and ``--export``.
        """

        args_parser = argparse.ArgumentParser()

        args_parser.add_argument(
            "--config", "-c",
            required=True,
            metavar="FILE",
            help="Path to plugin's YAML config file."
        )

        args_parser.add_argument(
            "--export", "-e",
            required=True,
            metavar="FILE",
            help="Path to the output file. Use '-' for standard output instead of a file."
        )

        return args_parser

    def parse_args(self) -> argparse.Namespace:
        """
        Parse command-line arguments.

        Common options ``--config`` and ``--export`` are consumed immediately.
        """

        parser = self.create_args_parser()

        options = parser.parse_args()

        self.config_filepath = options.config
        self.export_filepath = options.export

        with open(options.config) as f:
            self.config = ruamel.yaml.YAML().load(f)

        return options

    def fetch_data(self) -> None:
        """
        This method is responsible for acquiring data the exporter needs to emit metrics.
        """

        pass

    def create_metric(self) -> Iterator[Metric]:
        """
        This method is expected to yield one metric every time is called, or return when exporter
        runs out of metrics to export.
        """

        # This method is supposed to be a generator, overriden by child classes. We want to provide a default
        # behavior, which would be an "empty" iterator - when called, this method should yield no entries,
        # iteration over it should end immediately. This is achieved by doing nothing and calling `return`
        # right away.
        #
        # But `return` alone is not enough - this function must be a generator, and generator must have `yield`.
        # So, adding `yield` - and the order is important, `return` must go first. Othwerwise, `yield` would
        # actually yielded an item, `None`, but as said above, this generator is supposed to be empty.
        #
        # So, `return` to quit the generator, and `yield` to "trick" Python to treating this method as
        # a generator. The `yield` is never used, all calls end with `return`, but it's present.
        return
        yield

    def export_prometheus_data(self) -> None:
        """
        Main workhorse: calls ``create_metric`` as long as necessary, adds gained metrics to a registry,
        and emits the output.
        """

        registry = CollectorRegistry()

        for metric in self.create_metric():
            registry.register(metric)

        exported = generate_latest(registry)

        if self.export_filepath == '-':
            sys.stdout.write(exported.decode('utf-8'))

        else:
            with open(self.export_filepath, 'w') as f:
                f.write(exported.decode('utf-8'))

    def run(self) -> None:
        self.parse_args()

        try:
            self.fetch_data()

        except Exception as exc:
            print('Failed to fetch data: {}'.format(exc), file=sys.stderr)
            print(file=sys.stderr)
            traceback.print_exc(file=sys.stderr)

            sys.exit(1)

        try:
            self.export_prometheus_data()

        except Exception as exc:
            print('Failed export metrics: {}'.format(exc), file=sys.stderr)
            print(file=sys.stderr)
            traceback.print_exc(file=sys.stderr)

            sys.exit(1)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['tft', 'tft.phoebe']

package_data = \
{'': ['*']}

install_requires = \
['beautifulsoup4>=4.8.1,<5.0.0',
 'lxml>=4.4.2,<5.0.0',
 'prometheus-client>=0.7.1,<0.8.0',
 'requests',
 'ruamel.yaml>=0.15.51,<0.16.0']

entry_points = \
{'console_scripts': ['phoebe-jenkins = tft.phoebe.exporter_jenkins:main',
                     'phoebe-openstack = tft.phoebe.exporter_openstack:main']}

setup_kwargs = {
    'name': 'tft-phoebe',
    'version': '0.0.2',
    'description': 'Data gathering proxy for Prometheus toolkit',
    'long_description': "= Phoebe\n\nPhoebe is a data gathering proxy for https://prometheus.io/[Prometheus] toolkit. Its goal is to gather metrics from systems that don't provide their own native Prometheus export, and make them available for digestion by Prometheus.\n\n\n== How to run an exporter\n\n[source,shell]\n....\n$ pip install tft-phoebe\n$ phoebe-openstack --config=<your config file> --export=<filepath or - for stdout>\n....\n\nOr, after checking out the sources:\n\n[source,shell]\n....\n$ poetry install\n$ poetry run phoebe-openstack --config=<your config file> --export=<filepath or - for stdout>\n....\n",
    'author': 'Jakub Haruda',
    'author_email': 'jharuda@redhat.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://pypi.org/project/tft-phoebe',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
